# These are the default "django-admin startproject" settings.py files
# for each supported version of Django, with:
#
#   * "anymail" appended to INSTALLED_APPS
#   * ROOT_URLCONF = 'tests.test_settings.urls'
#
# Keeping a file for each Django version is simpler than trying
# to maintain warning-free compatibility with all Django versions
# in a single settings list.
#
# It also helps ensure compatibility with default apps, middleware, etc.
#
# You can typically find the default settings.py template in
# SITE_PACKAGES/django/conf/project_template/project_name/settings.py-tpl
